# coding=utf-8


# -------------------------------------------------
# PANGAEA
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from tools.taxon import Taxon
from shutil import copyfile


class PANGAEA(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__db = model.subsystem_getDb()

        # Db
        self.__conn = None
        self.__cursor = None

        # Puffer
        self.__bPufferHasUnsavedData = False  # Besitzt der Puffer auf-HDD-ungespeicherte Daten?
        self.__sFileNamePufferPANGAEA = "pangaea.dat"
        # Campaign
        self.__dicPufferPANGAEA_campaign_found = {}
        self.__dicPufferPANGAEA_campaign_notfound = {}
        # Event
        self.__dicPufferPANGAEA_event_found = {}
        self.__dicPufferPANGAEA_event_notfound = set()
        # Statistik
        self.__odicPufferPANGAEA_ageGroup_campaign_notfound = OrderedDict()
        self.__odicPufferPANGAEA_ageGroup_event_notfound = OrderedDict()
        self.__fPufferPANGAEA_meanAge_campaign_notfound = 0
        self.__fPufferPANGAEA_meanAge_event_notfound = 0

    # --- Info ---

    def info(self):
        sOut = "*** Status of myPANGAEA\n\n"
        if self.puffer_getSize() > 0:
            sOut += self.puffer_getStatusString() + "\n"
        return sOut

    # --- Admin ---

    def puffer_hasUnsavedData(self):
        return self.__bPufferHasUnsavedData

    def puffer_getSize(self):
        iSize = len(self.__dicPufferPANGAEA_event_found) + len(self.__dicPufferPANGAEA_event_notfound) + len(self.__dicPufferPANGAEA_campaign_found) + len(self.__dicPufferPANGAEA_campaign_notfound)
        return iSize

    def puffer_getSize_campaign(self):
        iCampaign_notfound = len(self.__dicPufferPANGAEA_campaign_notfound)
        iCampaign_found = len(self.__dicPufferPANGAEA_campaign_found)
        iCampaign_total = iCampaign_found + iCampaign_notfound
        fCampaign_notfound_ratio = 0
        try:
            fCampaign_notfound_ratio = iCampaign_notfound / iCampaign_total
        except:
            pass
        return iCampaign_total, iCampaign_notfound, iCampaign_found, fCampaign_notfound_ratio
        
    def puffer_getSize_event(self):
        iEvent_notfound = len(self.__dicPufferPANGAEA_event_notfound)
        iEvent_found = len(self.__dicPufferPANGAEA_event_found)
        iEvent_total = iEvent_found + iEvent_notfound
        fEvent_notfound_ratio = 0
        try:
            fEvent_notfound_ratio = iEvent_notfound / iEvent_total
        except:
            pass
        return iEvent_total, iEvent_notfound, iEvent_found, fEvent_notfound_ratio

    def puffer_cleanDisk(self, **kwargs):
        self.puffer_cleanRam(**kwargs)
        self.puffer_save()

    def puffer_cleanRam(self, **kwargs):
        # Parameter
        if "cruise" in kwargs or "events" in kwargs:
            bCruise = kwargs["cruise"] if "cruise" in kwargs else False
            bEvents = kwargs["events"] if "events" in kwargs else False
        else:
            bCruise = True
            bEvents = True
        # Taxa
        if bCruise:
            self.__dicPufferPANGAEA_campaign_found = {}
            self.__dicPufferPANGAEA_campaign_notfound = {}
            self.__bPufferHasUnsavedData = True
        # Name Search
        if bEvents:
            self.__dicPufferPANGAEA_event_found = {}
            self.__dicPufferPANGAEA_event_notfound = {}
            self.__bPufferHasUnsavedData = True

    def puffer_save(self):
        res = Result("Save myPANGAEA buffer")
        if self.puffer_hasUnsavedData():
            sFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferPANGAEA
            try:
                date = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
                data = (date, self.__dicPufferPANGAEA_event_found, self.__dicPufferPANGAEA_event_notfound, self.__dicPufferPANGAEA_campaign_found, self.__dicPufferPANGAEA_campaign_notfound)
                with open(sFile, 'wb') as handle:
                    pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
                    res.addData("file", sFile)
                    res.addData("date", date)
                self.__bPufferHasUnsavedData = False
            except Exception as ex:
                res.addErr("Problems saving buffer", add=str(ex), kat="I/O", sys="myPANGAEA")
        else:
            res.addInfo("Saving Buffer", add="Buffer has no unsaved data -> Saving buffer not needed ", kat="I/O",
                        sys="myPANGAEA")
            res.addData("save_not_needed", True)
        return res

    def puffer_import(self, sFile):
        pn = self.__model.ui_getSmartprinter().normal
        res = Result("Import myPANGAEA buffer")
        try:
            # Restore Backup
            if FileTools.getFileType(sFile) == "zip":
                bPANGAEArestoreFileZip = True
                sPANGAEArestoreFileZip = sFile
                sPANGAEArestoreFile = FileTools.getFileWithoutType(sPANGAEArestoreFileZip)
                sPANGAEArestoreFileName = FileTools.getFileWithoutPath(sPANGAEArestoreFile)
                sPANGAEArestoreFilePath = FileTools.getFilePath(sPANGAEArestoreFile)
                resCheckFileExists = FileTools.prepareSaveFile(sPANGAEArestoreFile)
                if resCheckFileExists.getData("bak") is not None:
                    pn("* Backup existing file")
                    pn(str(sFile) + " -> " + str(resCheckFileExists.getData("bak")))
                if sPANGAEArestoreFile in ["", None] or sPANGAEArestoreFileZip in ["", None]:
                    res.addErr("Problems unzipping file", add=str(sPANGAEArestoreFileZip), kat="Zip", sys="Admin")
                # Unzip
                pn("* Unzipping file")
                pn(str(sPANGAEArestoreFileZip))
                pn("->")
                pn(str(sPANGAEArestoreFile))
                lstZipContent = []
                try:
                    lstZipContent = zipfile.ZipFile(sPANGAEArestoreFileZip).namelist()
                    zipfile.ZipFile(sPANGAEArestoreFileZip).extract(sPANGAEArestoreFileName,
                                                                    path=sPANGAEArestoreFilePath)
                    pn("done")
                except Exception as ex:
                    res.addErr("File not found in zip package",
                               add=str(sPANGAEArestoreFileName) + " -> found only this files: " + PrettyText.lst2strC(
                                   lstZipContent), kat="Zip", sys="Admin")
                    pn("error")
            else:
                bPANGAEArestoreFileZip = False
                sPANGAEArestoreFileZip = None
            pass

            # Alten Puffer löschen
            self.puffer_cleanRam()

            # Puffer kopieren
            sPANGAEA_buffer = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferPANGAEA
            if bPANGAEArestoreFileZip:
                copyfile(sPANGAEArestoreFile, sPANGAEA_buffer)
            else:
                copyfile(sFile, sPANGAEA_buffer)

            # Die enpackte Datei loeschen
            if bPANGAEArestoreFileZip:
                pn("* Removing unzipped file")
                pn(str(sPANGAEArestoreFile))
                os.remove(sPANGAEArestoreFile)
                pn("done")

        except Exception as ex:
            res.addErr("Problems saving copy of myPANGAEA buffer", add=str(ex), kat="I/O", sys="myPANGAEA")
        return res

    def puffer_export(self, sFile):
        res = Result("Export myPANGAEA buffer")
        try:
            # Puffer kopieren
            sPANGAEA = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferPANGAEA
            sFileName = FileTools.getFileWithoutPath(sFile)
            sFilePath = FileTools.getFilePath(sFile)

            copyfile(sPANGAEA, sFile)
            # Kopie zippen und Original löschen
            zipfile.ZipFile(sFile + ".zip", mode='w', compression=zipfile.ZIP_DEFLATED).write(sFile, sFileName)
            os.remove(sFile)
            res.addData("file", str(sFileName) + ".zip")
            res.addData("folder", str(sFilePath))
        except Exception as ex:
            res.addErr("Problems saving copy of myPANGAEA buffer", add=str(ex), kat="I/O", sys="myPANGAEA")
        return res

    def puffer_snapshot(self):
        res = Result("Snapshot myPANGAEA buffer")
        sOrigFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferPANGAEA
        sSnapshotFolder = self.__model.conf_getAppSnapshotDirPANGAEA()
        sSnapshotFileName = "pangaea_" + PrettyText.createTimestamp() + ".dat"
        sSnapshotFile = sSnapshotFolder + "/" + sSnapshotFileName
        try:
            res.addData("snapshot-filename", sSnapshotFileName + ".zip")
            res.addData("snapshot-folder", sSnapshotFolder)
            res.addData("snapshot-file", sSnapshotFile)
            # Puffer kopieren
            copyfile(sOrigFile, sSnapshotFile)
            # Kopie zippen und Original löschen
            zipfile.ZipFile(sSnapshotFile + ".zip", mode='w', compression=zipfile.ZIP_DEFLATED).write(sSnapshotFile,sSnapshotFileName)
            os.remove(sSnapshotFile)
        except Exception as ex:
            res.addErr("Problems Taking Buffer Snapshot", add=str(ex), kat="I/O", sys="myPANGAEA")
        return res

    def puffer_load(self):
        res = Result("Reload myPANGAEA buffer")
        sFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferPANGAEA
        try:
            with open(sFile, 'rb') as handle:
                self.puffer_cleanRam()
                date, self.__dicPufferPANGAEA_event_found, self.__dicPufferPANGAEA_event_notfound, self.__dicPufferPANGAEA_campaign_found, self.__dicPufferPANGAEA_campaign_notfound  = pickle.load(handle)
                resStat = self.puffer_createStats()
                if resStat.isErr():
                    res.addResult(resStat, msg_only=False)
                res.addData("file", sFile)
                res.addData("date", date)
                # # Scatter
                # for iAid, t in self.__dicPufferWorms.items():
                #     datum = t.getDateCreate()
                #     t.debug__setDateCreate(datum - datetime.timedelta(days=random.randint(1,90)))
                #     self.__dicPufferWorms[iAid] = t
        except Exception as ex:
            self.puffer_cleanDisk()
            res.addErr("Problems loading buffer -> Made a new empty buffer", add=str(ex), kat="I/O", sys="myPANGAEA")
        return res

    # --- Work ---

    def checkCampaign(self, sCampaign, **kwargs):
        return self.__check("campaign", sCampaign, **kwargs)

    def checkEvent(self, sEvent, **kwargs):
        return self.__check("event", sEvent, **kwargs)

    def __check(self, sSystem, sName, **kwargs):
        mypangaeaUpdate = self.__smartprinter.updateMyPANGAEA
        res = Result(f"Check myPANGAEA: {sSystem}")

        # System
        if sSystem == "event":
            dicFound = self.__dicPufferPANGAEA_event_found
            dicFound_not = self.__dicPufferPANGAEA_event_notfound
            fkt = self.__db.pangaea_checkEvent
        elif sSystem == "campaign":
            dicFound = self.__dicPufferPANGAEA_campaign_found
            dicFound_not = self.__dicPufferPANGAEA_campaign_notfound
            fkt = self.__db.pangaea_checkCampaign
        else:
            res.addErr("Internal problem", add=f"System unknown: {sSystem}", kat="buffer", sys="myPANGAEA")
            return res

        # Parameter
        bReask_known  = kwargs["reask_known"] if "reask_known" in kwargs else False
        bReask_unknown  = kwargs["reask_unknown"] if "reask_unknown" in kwargs else False

        # Puffer
        bPuffer = False
        bFound = False
        if not bReask_known and sName in dicFound:
            bPuffer = True
            bFound = True
        if not bReask_unknown and sName in dicFound_not:
            bPuffer = True
            bFound = False
        if bPuffer:
            # Gefunden in Puffer
            res.addData("found", bFound)
            res.addData("puffer", True)
        else:
            # Ask PANGAEA
            res.addData("puffer", False)
            res_ask = fkt(sName)
            if not res_ask.isErr():
                self.__bPufferHasUnsavedData = True
                bFound = res_ask.getData("found")
                if bFound:
                    dicFound[sName] = datetime.datetime.now()
                else:
                    dicFound_not[sName] = datetime.datetime.now()
                mypangaeaUpdate() # Update UI/CLI
            res.addResult(res_ask)
        return res

    def puffer_createStats(self):
        res = Result("myPANGAEA: Create statistics")

        # Helper
        def createStat(dicData):
            lstDays = []
            dicAgeGroup = {}
            fMeanAge = 0
            # Age Groups
            for sName, dateEvent in dicData.items():
                iDays = (dateNow - dateEvent).days
                lstDays.append(iDays)
                if iDays not in dicAgeGroup:
                    dicAgeGroup[iDays] = []
                dicAgeGroup[iDays].append(sName)
            dicAgeGroup = OrderedDict(sorted(dicAgeGroup.items(), reverse=True))
            # Mittleres Alter
            if len(lstDays) != 0:
                fMeanAge = sum(lstDays) / len(lstDays)
            return dicAgeGroup, fMeanAge

        # --- Alle not-found Events/Campaigns nach Alter in Tagen gruppieren
        dateNow = datetime.datetime.now()
        try:
            # Campaign
            self.__odicPufferPANGAEA_ageGroup_campaign_notfound, self.__fPufferPANGAEA_meanAge_campaign_notfound = createStat(self.__dicPufferPANGAEA_campaign_notfound)
            # Event
            self.__odicPufferPANGAEA_ageGroup_event_notfound, self.__fPufferPANGAEA_meanAge_event_notfound = createStat(self.__dicPufferPANGAEA_event_notfound)
        except Exception as ex:
            res.addErr("Creating statistics", add=str(ex), kat="buffer", sys="myPANGAEA")
        return res

    def puffer_getStatusString(self, **kwargs):
        # Parameter
        iCampaign_found = len(self.__dicPufferPANGAEA_campaign_found)
        iCampaign_notfound = len(self.__dicPufferPANGAEA_campaign_notfound)
        iEvent_found = len(self.__dicPufferPANGAEA_event_found)
        iEvent_notfound = len(self.__dicPufferPANGAEA_event_notfound)

        # Info

        # Overview
        sOut = ""
        sOut += "--- Buffered Data ---\n\n"
        lstTable = [["Type", "PANGAEA", "Number"]]
        lstTable.append(["Campaign", "known", iCampaign_found])
        lstTable.append(["", "unknown", iCampaign_notfound])
        lstTable.append(["Event", "known", iEvent_found])
        lstTable.append(["", "unknown", iEvent_notfound])
        sOut += PrettyText.table(lstTable,True)
        sOut += "\n\n"

        # Age groups
        # Campaign
        if iCampaign_notfound > 0:
            lstBufferStatus = [["Age", "Campaigns"]]
            # Bar Plot
            lstBarchartData = [("Age Group", "Number")]
            for iDay, lstEvents in self.__odicPufferPANGAEA_ageGroup_campaign_notfound.items():
                lstBarchartData.append((str(iDay) + " days", len(lstEvents)))

            sOut += "--- Unknown campaigns ---\n\n"
            iCampaign_total, iCamaign_notfound, iCampaign_found, fCampaign_notfound_ratio = self.puffer_getSize_campaign()
            sOut += PrettyText.barchart(lstBarchartData, maxbarsize=25, header=True) + "\n"
            sOut += f"Median age: {self.__fPufferPANGAEA_meanAge_campaign_notfound:.1f} days\n\n"

        # Event
        if iEvent_notfound > 0:
            lstBufferStatus = [["Age", "Events"]]
            # Bar Plot
            lstBarchartData = [("Age Group", "Number")]
            for iDay, lstEvents in self.__odicPufferPANGAEA_ageGroup_event_notfound.items():
                lstBarchartData.append((str(iDay) + " days", len(lstEvents)))

            sOut += "--- Unknown events ---\n\n"
            iEvent_total, iEvent_notfound, iEvent_found, fEvent_notfound_ratio = self.puffer_getSize_event()
            sOut += PrettyText.barchart(lstBarchartData, maxbarsize=25, header=True) + "\n"
            sOut += f"Median age: {self.__fPufferPANGAEA_meanAge_event_notfound:.1f} days\n"

        return sOut

    def puffer_update(self, **kwargs):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status
        mypangaeaUpdate = self.__smartprinter.updateMyPANGAEA

        res = Result(name="Update myPANGAEA buffer")

        # Init
        self.puffer_cleanRam()
        self.puffer_load()
        resStat = self.puffer_createStats()
        if resStat.isErr():
            res.addResult(resStat, msg_only=False)
        self.abort(False)

        # Log
        progress("on-pulse-abort")
        sLog = ""
        sLog += "************************************\n"
        sLog += "Update myPANGAEA buffer\n"
        sLog += "************************************\n\n"

        # Startzeit
        iZeitStart = time.time()
        sZeitStart = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))

        # Data
        lstBuffers = [
            ("Campaign", self.__dicPufferPANGAEA_campaign_notfound, self.__odicPufferPANGAEA_ageGroup_campaign_notfound, self.checkCampaign),
            ("Event", self.__dicPufferPANGAEA_event_notfound, self.__odicPufferPANGAEA_ageGroup_event_notfound, self.checkEvent)
        ]

        # Loop Data
        iNr_total_now = 0
        iNr_total_max = len(self.__dicPufferPANGAEA_campaign_notfound) + len(self.__dicPufferPANGAEA_event_notfound)
        iFound = 0
        for (sName, odicPuffer, odicAge, checkFkt) in lstBuffers:
            pv(f"* Updating unknown {sName}s", fmt=False)

            iAgeGroups = len(odicAge.keys())
            iAgeGroup = 0
            iNr_system_total = len(odicPuffer)
            lstStat = [["Age group", "number", "found in PANGAEA"]]
            sLog += f"--- Unknown {sName} ---\n\n"

            # Loop: Age Groups
            for iDay_group, lstData in odicAge.items():
                if self.isAbort():
                    break
                iAgeGroup += 1
                iDayGroup_found = 0
                iDayGroup_nr_now = 0
                iDayGroup_nr_max = len(lstData)
                workflow(top="myPANGAEA", sub=f"Updating unknown {sName}s", fmt=False)
                if self.isAbort():
                    break
                # Loop in Age Group
                for sData in lstData:
                    if self.isAbort():
                        break
                    # Update
                    iNr_total_now += 1
                    iDayGroup_nr_now += 1
                    sText = f"Age group ({iAgeGroup}/{iAgeGroups}): {iDay_group}  days:"
                    sStatus = f"Updating {sName}: {sData}"
                    progress(iNr_total_now, iNr_total_max, 100, proz=True, abs=True, cli=True, text=sText, fmt_ascii_only=False)
                    status(sStatus)
                    # Ask PANGAEA
                    resCheck = checkFkt(sData, reask_unknown=True)
                    bFound = resCheck.getData("found")
                    bBuffer = resCheck.getData("puffer")
                    sInfo = f"{sData}: "
                    if bFound:
                        # Gefunden
                        sInfo += "found!"
                        iFound+=1
                        iDayGroup_found+=1
                        mypangaeaUpdate()
                    else:
                        # Nicht gefunden
                        sInfo += "-"
                    pv(sInfo)

                # Log
                lstStat.append([f"{iDay_group} days", iDayGroup_nr_max, iDayGroup_found if iDayGroup_found > 0 else "-"])

            lstStat.append(["total", iNr_system_total, iFound if iFound > 0 else "-"])

            sLog += PrettyText.table(lstStat, True)
            sLog +="\n\n"

        # Log Zeit
        iSekTotal = time.time() - iZeitStart
        iStd = int(iSekTotal // 3600)
        iMin = int((iSekTotal % 3600) // 60)
        iSek = int(iSekTotal % 60)

        sLog += "--- Time ---\n\n"
        sLog += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLog += "\n"

        if self.isAbort():
            sLog += "\n<Stopped by user>\n\n"

        # Puffer save an Rebuild Stats
        self.puffer_save()
        resStat = self.puffer_createStats()
        if resStat.isErr():
            res.addResult(resStat, msg_only=False)

        res.addData("title", "Update of myPANGAEA buffer")
        res.addData("text", sLog)
        progress("off")
        return res

